# -*- coding: utf-8 -*-

# built ins
import asyncio
import json
from datetime import datetime, timedelta, timezone

# installed
import orjson
import uvloop
import websockets
from dataclassy import dataclass, fields

asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

# user defined formula
from streaming_helper.restful_api.deribit import (
    end_point_params_template,
    end_point_template,
)
from streaming_helper.utilities import error_handling, string_modification as str_mod


@dataclass(unsafe_hash=True, slots=True)
class StreamingAccountData:
    """ """

    sub_account_id: str
    client_id: str
    client_secret: str
    # Async Event Loop
    loop = asyncio.get_event_loop()
    ws_connection_url: str = "wss://www.deribit.com/ws/api/v2"
    # Instance Variables
    websocket_client: websockets.WebSocketClientProtocol = None
    refresh_token: str = None
    refresh_token_expiry_time: int = None

    async def ws_manager(
        self,
        client_redis,
        exchange,
        queue_general: object,
        futures_instruments,
        resolutions: list,
    ) -> None:

        async with websockets.connect(
            self.ws_connection_url,
            ping_interval=None,
            compression=None,
            close_timeout=60,
        ) as self.websocket_client:

            try:

                instruments_name = futures_instruments["instruments_name"]

                while True:

                    # Authenticate WebSocket Connection
                    await self.ws_auth()

                    # Establish Heartbeat
                    await self.establish_heartbeat()

                    # Start Authentication Refresh Task
                    self.loop.create_task(self.ws_refresh_auth())

                    ws_instruments = []

                    instrument_kinds = ["future, future_combo"]

                    for kind in instrument_kinds:

                        user_changes = f"user.changes.{kind}.any.raw"
                        ws_instruments.append(user_changes)

                    orders = f"user.orders.any.any.raw"
                    ws_instruments.append(orders)
                    trades = f"user.trades.any.any.raw"
                    ws_instruments.append(trades)

                    for instrument in instruments_name:

                        if "PERPETUAL" in instrument:

                            currency = str_mod.extract_currency_from_text(instrument)
                            portfolio = f"user.portfolio.{currency}"

                            ws_instruments.append(portfolio)

                            for resolution in resolutions:

                                ws_chart = f"chart.trades.{instrument}.{resolution}"
                                ws_instruments.append(ws_chart)

                        incremental_ticker = f"incremental_ticker.{instrument}"

                        ws_instruments.append(incremental_ticker)

                    await self.ws_operation(
                        operation="subscribe",
                        ws_channel=ws_instruments,
                        source="ws-combination",
                    )

                    while True:

                        # Receive WebSocket messages
                        message: bytes = await self.websocket_client.recv()
                        message: dict = orjson.loads(message)

                        if "id" in list(message):
                            if message["id"] == 9929:

                                if self.refresh_token is None:
                                    print(
                                        "Successfully authenticated WebSocket Connection"
                                    )

                                else:
                                    print(
                                        "Successfully refreshed the authentication of the WebSocket Connection"
                                    )

                                self.refresh_token = message["result"]["refresh_token"]

                                # Refresh Authentication well before the required datetime
                                if message["testnet"]:
                                    expires_in: int = 300
                                else:
                                    expires_in: int = (
                                        message["result"]["expires_in"] - 240
                                    )

                                now_utc: int = datetime.now(timezone.utc)

                                self.refresh_token_expiry_time = now_utc + timedelta(
                                    seconds=expires_in
                                )

                            elif message["id"] == 8212:
                                # Avoid logging Heartbeat messages
                                continue

                        elif "method" in list(message):
                            # Respond to Heartbeat Message
                            if message["method"] == "heartbeat":
                                await self.heartbeat_response()

                        if "params" in list(message):

                            if message["method"] != "heartbeat":

                                message_params: dict = message["params"]

                                if message_params:

                                    message_params.update({"exchange": exchange})
                                    message_params.update(
                                        {"account_id": self.sub_account_id}
                                    )

                                    # queing message to dispatcher
                                    await queue_general.put(message_params)

                                """
                                    message examples:
                                    
                                    incremental_ticker = {
                                        'channel': 'incremental_ticker.BTC-7FEB25', 
                                        'data': {
                                            'timestamp': 1738407481107, 
                                            'type': 'snapshot', 
                                            'state': 'open',
                                            'stats': {
                                                'high': 106245.0, 
                                                'low': 101550.0, 
                                                'price_change': -2.6516, 
                                                'volume': 107.12364526, 
                                                'volume_usd': 11081110.0, 
                                                'volume_notional': 11081110.0
                                                }, 
                                                'index_price': 101645.32,
                                                'instrument_name': 'BTC-7FEB25', 
                                                'last_price': 101787.5, 
                                                'settlement_price': 102285.5, 
                                                'min_price': 100252.5,
                                                'max_price': 103310.0,
                                                'open_interest': 18836380, 
                                                'mark_price': 101781.52,
                                                'best_ask_price': 101780.0, 
                                                'best_bid_price': 101775.0,
                                                'estimated_delivery_price': 101645.32,
                                                'best_ask_amount': 15500.0, 
                                                'best_bid_amount': 11310.0
                                                }
                                                }
                                            
                                    chart.trades = {
                                        'channel': 'chart.trades.BTC-PERPETUAL.1', 
                                        'data': {
                                            'close': 101650.5, 
                                            'high': 101650.5, 
                                            'low': 101650.5, 
                                            'open': 101650.5, 
                                            'tick': 1738407480000, 
                                            'cost': 0.0, 
                                            'volume': 0.0}
                                            }
                                            
                                    portfolio = {
                                        'channel': 'user.portfolio.btc', 
                                        'data': {
                                            'options_pl': 0.0, 
                                            'balance': 0.00214241, 
                                            'session_rpl': 0.0,
                                            'initial_margin': 0.00075353, 
                                            'additional_reserve': 0.0, 
                                            'spot_reserve': 0.0, 
                                            'futures_pl': -1.442e-05,
                                            'total_delta_total_usd': 193.408186282, 
                                            'total_maintenance_margin_usd': 53.465166986729, 
                                            'options_theta_map': {}, 
                                            'projected_delta_total': 0.002373, 
                                            'options_gamma': 0.0, 
                                            'total_pl': -1.442e-05, 
                                            'options_gamma_map': {},
                                            'projected_initial_margin': 0.00075353, 
                                            'options_session_rpl': 0.0,
                                            'options_session_upl': 0.0, 
                                            'total_margin_balance_usd': 273.683871785, 
                                            'equity': 0.00213253, 
                                            'cross_collateral_enabled': True, 
                                            'delta_total': 0.002373, 
                                            'delta_total_map': {
                                                'btc_usd': 0.002373282}, 
                                                'options_value': 0.0, 
                                                'total_equity_usd': 273.683871785,
                                                'locked_balance': 0.0, 
                                                'margin_balance': 0.00269254, 
                                                'available_withdrawal_funds': 0.00203902,
                                                'options_delta': 0.0,
                                                'currency': 'BTC', 
                                                'fee_balance': 0.0,
                                                'options_vega_map': {}, 
                                                'available_funds': 0.00193902, 
                                                'maintenance_margin': 0.000526, 
                                                'futures_session_rpl': 0.0, 
                                                'total_initial_margin_usd': 76.592212527, 
                                                'session_upl': -9.88e-06, 
                                                'options_vega': 0.0, 
                                                'options_theta': 0.0,
                                                'futures_session_upl': -9.88e-06,
                                                'portfolio_margining_enabled': True, 
                                                'projected_maintenance_margin': 0.000526, 
                                                'margin_model': 'cross_pm'
                                                }
                                                }
                                    """

            except Exception as error:
                
                await error_handling.parse_error_message_with_redis(
            client_redis,
            error,
        )


    async def establish_heartbeat(self) -> None:
        """
        reference: https://github.com/ElliotP123/crypto-exchange-code-samples/blob/master/deribit/websockets/dbt-ws-authenticated-example.py

        Requests DBT's `public/set_heartbeat` to
        establish a heartbeat connection.
        """
        msg: dict = {
            "jsonrpc": "2.0",
            "id": 9098,
            "method": "public/set_heartbeat",
            "params": {"interval": 10},
        }

        try:
            await self.websocket_client.send(json.dumps(msg))

        except Exception as error:

            await error_handling.parse_error_message(
                client_redis,
            )

    async def heartbeat_response(self) -> None:
        """
        Sends the required WebSocket response to
        the Deribit API Heartbeat message.
        """
        msg: dict = {
            "jsonrpc": "2.0",
            "id": 8212,
            "method": "public/test",
            "params": {},
        }

        try:

            await self.websocket_client.send(json.dumps(msg))

        except Exception as error:

            error_handling.parse_error_message(error)

    async def ws_auth(self) -> None:
        """
        Requests DBT's `public/auth` to
        authenticate the WebSocket Connection.
        """
        msg: dict = {
            "jsonrpc": "2.0",
            "id": 9929,
            "method": "public/auth",
            "params": {
                "grant_type": "client_credentials",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            },
        }

        try:
            await self.websocket_client.send(json.dumps(msg))

        except Exception as error:

            error_handling.parse_error_message(error)

    async def ws_refresh_auth(self) -> None:
        """
        Requests DBT's `public/auth` to refresh
        the WebSocket Connection's authentication.
        """
        while True:

            now_utc = datetime.now(timezone.utc)

            if self.refresh_token_expiry_time is not None:

                if now_utc > self.refresh_token_expiry_time:

                    msg: dict = {
                        "jsonrpc": "2.0",
                        "id": 9929,
                        "method": "public/auth",
                        "params": {
                            "grant_type": "refresh_token",
                            "refresh_token": self.refresh_token,
                        },
                    }

                    await self.websocket_client.send(json.dumps(msg))

            await asyncio.sleep(150)

    async def ws_operation(
        self,
        operation: str,
        ws_channel: str,
        source: str = "ws-single",
    ) -> None:
        """
        Requests `public/subscribe` or `public/unsubscribe`
        to DBT's API for the specific WebSocket Channel.

        source:
        ws-single
        ws-combination
        rest
        """
        sleep_time: int = 0.05

        await asyncio.sleep(sleep_time)

        id = end_point_template.id_numbering(
            operation,
            ws_channel,
        )

        msg: dict = {
            "jsonrpc": "2.0",
        }

        if "ws" in source:

            if "single" in source:
                ws_channel = [ws_channel]

            extra_params: dict = dict(
                id=id,
                method=f"private/{operation}",
                params={"channels": ws_channel},
            )

            msg.update(extra_params)

            if msg["params"]["channels"]:
                await self.websocket_client.send(json.dumps(msg))

        if "rest_api" in source:

            extra_params: dict = await end_point_params_template.get_api_end_point(
                operation,
                ws_channel,
            )

            msg.update(extra_params)

            await self.websocket_client.send(json.dumps(msg))



# -*- coding: utf-8 -*-


def extract_currency_from_text(words: str) -> str:
    """

    some variables:
    chart.trades.BTC-PERPETUAL.1
    incremental_ticker.BTC-4OCT24
    """

    if "." in words:
        filter1 = (words.partition(".")[2]).lower()

        if "." in filter1:
            filter1 = (filter1.partition(".")[2]).lower()

            if "chart.trades" in words:
                filter1 = (words.partition(".")[2]).lower()

            if "." in filter1:
                filter1 = (filter1.partition(".")[2]).lower()

                if "." in filter1:
                    filter1 = (filter1.partition(".")[0]).lower()

    else:
        filter1 = (words.partition(".")[0]).lower()

    return (filter1.partition("-")[0]).lower()


def message_template() -> str:
    """ """

    result = {}
    result.update({"params": {}})
    result.update({"method": "subscription"})
    result["params"].update({"data": None})
    result["params"].update({"channel": None})
    result["params"].update({"stream": None})

    return result


def extract_integers_from_text(words: list) -> int:
    """
    Extracting integers from label text. More general than get integer in parsing label function
    """

    words_to_str = str(
        words
    )  # ensuring if integer used as argument, will be returned as itself

    return int("".join([o for o in words_to_str if o.isdigit()]))


def remove_double_brackets_in_list(data: list) -> list:
    """_summary_

    Args:
        data (list): instance: [
                                ['BTC-30AUG24', 'BTC-6SEP24', 'BTC-27SEP24', 'BTC-27DEC24',
                                'BTC-28MAR25', 'BTC-27JUN25', 'BTC-PERPETUAL'
                                ],
                                ['ETH-30AUG24', 'ETH-6SEP24', 'ETH-27SEP24', 'ETH-27DEC24',
                                'ETH-28MAR25', 'ETH-27JUN25', 'ETH-PERPETUAL'
                                ]
                                ]

    Returns:
        list: _description_

    Reference:
        https://stackoverflow.com/questions/952914/how-do-i-make-a-flat-list-out-of-a-list-of-lists
    """
    return [o for os in data for o in os]


def remove_apostrophes_from_json(json_load: list) -> int:
    """ """
    import ast

    return [ast.literal_eval(str(i)) for i in json_load]


def remove_redundant_elements(data: list) -> list:
    """
    Remove redundant items in a list

    Args:
        data (list)

    Returns:
        list:

    Example:
        data_original = ['A', 'A', 'B', 'B', 'B', 'C']
        data_cleaned = ['A','B','C']

    Reference:
        1. https://stackoverflow.com/questions/9427163/remove-duplicate-dict-in-list-in-python
        2. https://python.plainenglish.io/how-to-remove-duplicate-elements-from-lists-without-using-sets-in-python-5796e93e6d43
    """

    # Create an empty list
    result = []

    # Check if the data is a list and not empty
    if isinstance(data, list) and data != []:
        try:
            # Ref 1
            result = list({frozenset(item.items()): item for item in data}.values())

        except:
            # Ref 2
            result = list(dict.fromkeys(data))

    return result



# -*- coding: utf-8 -*-

# built ins
import asyncio

import uvloop
from loguru import logger as log

asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

from streaming_helper.db_management import redis_client, sqlite_management as db_mgt
from streaming_helper.restful_api.deribit import end_point_params_template as end_point
from streaming_helper.restful_api import connector
from streaming_helper.data_announcer.deribit import (
    get_instrument_summary,
    allocating_ohlc,
)
from streaming_helper.utilities import (
    caching,
    error_handling,
    pickling,
    string_modification as str_mod,
    system_tools,
)


async def caching_distributing_data(
    client_redis: object,
    currencies: list,
    initial_data_subaccount: dict,
    redis_channels: list,
    redis_keys: list,
    strategy_attributes,
    queue_general: object,
) -> None:
    """
    my_trades_channel:
    + send messages that "high probabilities" trade DB has changed
        sender: redis publisher + sqlite insert, update & delete
    + updating trading cache at end user
        consumer: fut spread, hedging, cancelling
    + checking data integrity
        consumer: app data cleaning/size reconciliation

    sub_account_channel:
    update method: REST
    + send messages that sub_account has changed
        sender: deribit API module
    + updating sub account cache at end user
        consumer: fut spread, hedging, cancelling
    + checking data integrity
        consumer: app data cleaning/size reconciliation

    sending_order_channel:
    + send messages that an order has allowed to submit
        sender: fut spread, hedging
    + send order to deribit
        consumer: processing order

    is_order_allowed_channel:
    + send messages that data has reconciled each other and order could be processed
        sender: app data cleaning/size reconciliation, check double ids
    + processing order
        consumer: fut spread, hedging

    """

    try:

        # preparing redis connection
        pubsub = client_redis.pubsub()

        chart_low_high_tick_channel: str = redis_channels["chart_low_high_tick"]
        portfolio_channel: str = redis_channels["portfolio"]
        sub_account_cached_channel: str = redis_channels["sub_account_cache_updating"]
        sqlite_updating_channel: str = redis_channels["sqlite_record_updating"]
        order_update_channel: str = redis_channels["order_cache_updating"]
        my_trade_receiving_channel: str = redis_channels["my_trade_receiving"]
        my_trades_channel: str = redis_channels["my_trades_cache_updating"]
        abnormal_trading_notices: str = redis_channels["abnormal_trading_notices"]

        ticker_cached_channel: str = redis_channels["ticker_cache_updating"]

        # prepare channels placeholders
        channels = [
            order_update_channel,
            sqlite_updating_channel,
        ]

        # subscribe to channels
        [await pubsub.subscribe(o) for o in channels]

        server_time = 0

        portfolio = []

        settlement_periods = get_settlement_period(strategy_attributes)

        futures_instruments = await get_instrument_summary.get_futures_instruments(
            currencies, settlement_periods
        )

        instruments_name = futures_instruments["instruments_name"]

        ticker_all_cached = combining_ticker_data(instruments_name)

        sub_account_cached_params = initial_data_subaccount["params"]

        sub_account_cached = sub_account_cached_params["data"]

        orders_cached = sub_account_cached["orders_cached"]

        positions_cached = sub_account_cached["positions_cached"]

        query_trades = f"SELECT * FROM  v_trading_all_active"

        result = str_mod.message_template()

        while True:

            message_params: str = await queue_general.get()

            async with client_redis.pipeline() as pipe:

                try:

                    data: dict = message_params["data"]

                    message_channel: str = message_params["channel"]

                    currency: str = str_mod.extract_currency_from_text(message_channel)

                    currency_upper = currency.upper()

                    pub_message = dict(
                        data=data,
                        server_time=server_time,
                        currency_upper=currency_upper,
                        currency=currency,
                    )

                    if "user." in message_channel:

                        if "portfolio" in message_channel:

                            result["params"].update({"channel": portfolio_channel})
                            result["params"].update({"data": pub_message})

                            await updating_portfolio(
                                pipe,
                                portfolio,
                                portfolio_channel,
                                result,
                            )

                        elif "changes" in message_channel:

                            log.critical(message_channel)
                            log.warning(data)

                            await updating_sub_account(
                                client_redis,
                                orders_cached,
                                positions_cached,
                                query_trades,
                                data,
                                sub_account_cached_channel,
                            )

                        else:

                            log.critical(message_channel)
                            log.warning(data)

                            result["params"].update({"data": data})

                            if "trades" in message_channel:

                                await trades_in_message_channel(
                                    pipe,
                                    data,
                                    my_trade_receiving_channel,
                                    orders_cached,
                                    result,
                                )

                            if "order" in message_channel:

                                await order_in_message_channel(
                                    pipe,
                                    data,
                                    order_update_channel,
                                    orders_cached,
                                    result,
                                )

                            my_trades_active_all = (
                                await db_mgt.executing_query_with_return(query_trades)
                            )

                            result["params"].update({"channel": my_trades_channel})
                            result["params"].update({"data": my_trades_active_all})

                            await redis_client.publishing_result(
                                pipe,
                                my_trades_channel,
                                result,
                            )

                    instrument_name_future = (message_channel)[19:]
                    if (
                        message_channel
                        == f"incremental_ticker.{instrument_name_future}"
                    ):

                        await incremental_ticker_in_message_channel(
                            pipe,
                            currency,
                            data,
                            instrument_name_future,
                            result,
                            pub_message,
                            server_time,
                            ticker_all_cached,
                            ticker_cached_channel,
                        )

                    if "chart.trades" in message_channel:

                        await chart_trades_in_message_channel(
                            pipe,
                            chart_low_high_tick_channel,
                            message_channel,
                            pub_message,
                            result,
                        )

                except:

                    pass

                await pipe.execute()

    except Exception as error:

        await error_handling.parse_error_message_with_redis(
            client_redis,
            error,
        )


def compute_notional_value(
    index_price: float,
    equity: float,
) -> float:
    """ """
    return index_price * equity


def get_index(ticker: dict) -> float:

    try:

        index_price = ticker["index_price"]

    except:

        index_price = []

    if index_price == []:
        index_price = ticker["estimated_delivery_price"]

    return index_price


async def updating_portfolio(
    pipe: object,
    portfolio: list,
    portfolio_channel: str,
    result: dict,
) -> None:

    params = result["params"]

    data = params["data"]

    if portfolio == []:
        portfolio.append(data["data"])

    else:
        data_currency = data["data"]["currency"]
        portfolio_currency = [o for o in portfolio if data_currency in o["currency"]]

        if portfolio_currency:
            portfolio.remove(portfolio_currency[0])

        portfolio.append(data["data"])

    result["params"]["data"].update({"cached_portfolio": portfolio})

    await redis_client.publishing_result(
        pipe,
        portfolio_channel,
        result,
    )


def get_settlement_period(strategy_attributes: list) -> list:

    return str_mod.remove_redundant_elements(
        str_mod.remove_double_brackets_in_list(
            [o["settlement_period"] for o in strategy_attributes]
        )
    )


def combining_ticker_data(instruments_name: str) -> list:
    """_summary_
    https://blog.apify.com/python-cache-complete-guide/]
    https://medium.com/@jodielovesmaths/memoization-in-python-using-cache-36b676cb21ef
    data caching
    https://medium.com/@ryan_forrester_/python-return-statement-complete-guide-138c80bcfdc7

    Args:
        instrument_ticker (_type_): _description_

    Returns:
        _type_: _description_
    """

    result = []
    for instrument_name in instruments_name:

        result_instrument = reading_from_pkl_data("ticker", instrument_name)

        if result_instrument:
            result_instrument = result_instrument[0]

        else:

            basic_https_connection_url = end_point.basic_https()

            endpoint_tickers = end_point.get_tickers_end_point(instrument_name)

            result_instrument = await connector.get_connected(
                basic_https_connection_url,
                endpoint_tickers,
            )

        result.append(result_instrument)

    return result


def reading_from_pkl_data(
    end_point: str,
    currency: str,
    status: str = None,
) -> dict:
    """ """

    path: str = system_tools.provide_path_for_file(end_point, currency, status)
    return pickling.read_data(path)


async def trades_in_message_channel(
    pipe: object,
    data: list,
    my_trade_receiving_channel: str,
    orders_cached: list,
    result: dict,
) -> None:
    """

    #! result example

    [
        {
            'label': 'customShort-open-1743595398537',
            'timestamp': 1743595416236,
            'state': 'filled',
            'price': 1870.05,
            'direction': 'sell',
            'index_price': 1869.77,
            'profit_loss': 0.0,
            'instrument_name': 'ETH-PERPETUAL',
            'trade_seq': 175723279,
            'api': True,
            'mark_price': 1869.94,
            'amount': 1.0,
            'order_id': 'ETH-64159311162',
            'matching_id': None,
            'tick_direction': 0,
            'fee': 0.0,
            'mmp': False,
            'post_only': True,
            'reduce_only': False,
            'self_trade': False,
            'contracts': 1.0,
            'trade_id': 'ETH-242752309',
            'fee_currency': 'ETH',
            'order_type': 'limit',
            'risk_reducing': False,
            'liquidity': 'M'
        }
    ]

    """

    result["params"].update({"channel": my_trade_receiving_channel})

    await redis_client.publishing_result(
        pipe,
        my_trade_receiving_channel,
        result,
    )

    for trade in data:

        log.info(trade)

        caching.update_cached_orders(
            orders_cached,
            trade,
        )


async def order_in_message_channel(
    pipe: object,
    data: list,
    order_update_channel: str,
    orders_cached: list,
    result: dict,
) -> None:

    currency: str = str_mod.extract_currency_from_text(data["instrument_name"])

    caching.update_cached_orders(
        orders_cached,
        data,
    )

    data = dict(
        current_order=data,
        open_orders=orders_cached,
        currency=currency,
        currency_upper=currency.upper(),
    )

    result["params"].update({"channel": order_update_channel})
    result["params"].update({"data": data})

    await redis_client.publishing_result(
        pipe,
        order_update_channel,
        result,
    )


async def incremental_ticker_in_message_channel(
    pipe: object,
    currency: str,
    data: list,
    instrument_name_future: str,
    result: dict,
    pub_message: dict,
    server_time: int,
    ticker_all_cached: list,
    ticker_cached_channel: str,
) -> None:

    # extract server time from data
    current_server_time = (
        data["timestamp"] + server_time if server_time == 0 else data["timestamp"]
    )

    currency_upper = currency.upper()

    # updating current server time
    server_time = (
        current_server_time if server_time < current_server_time else server_time
    )

    pub_message.update({"instrument_name": instrument_name_future})
    pub_message.update({"currency_upper": currency_upper})

    for item in data:

        if "stats" not in item and "instrument_name" not in item and "type" not in item:
            [
                o
                for o in ticker_all_cached
                if instrument_name_future in o["instrument_name"]
            ][0][item] = data[item]

        if "stats" in item:

            data_orders_stat = data[item]

            for item in data_orders_stat:
                [
                    o
                    for o in ticker_all_cached
                    if instrument_name_future in o["instrument_name"]
                ][0]["stats"][item] = data_orders_stat[item]

    pub_message = dict(
        data=ticker_all_cached,
        server_time=server_time,
        instrument_name=instrument_name_future,
        currency_upper=currency_upper,
        currency=currency,
    )

    result["params"].update({"channel": ticker_cached_channel})
    result["params"].update({"data": pub_message})

    await redis_client.publishing_result(
        pipe,
        ticker_cached_channel,
        result,
    )
    if "PERPETUAL" in instrument_name_future:

        WHERE_FILTER_TICK: str = "tick"

        resolution = 1

        TABLE_OHLC1: str = f"ohlc{resolution}_{currency}_perp_json"

        await allocating_ohlc.inserting_open_interest(
            currency,
            WHERE_FILTER_TICK,
            TABLE_OHLC1,
            data,
        )


async def chart_trades_in_message_channel(
    pipe: object,
    chart_low_high_tick_channel: str,
    message_channel: str,
    pub_message: list,
    result: dict,
) -> None:

    try:
        resolution = int(message_channel.split(".")[3])

    except:
        resolution = message_channel.split(".")[3]

    pub_message.update({"instrument_name": message_channel.split(".")[2]})
    pub_message.update({"resolution": resolution})

    result["params"].update({"channel": chart_low_high_tick_channel})
    result["params"].update({"data": pub_message})

    await redis_client.publishing_result(
        pipe,
        chart_low_high_tick_channel,
        result,
    )


async def updating_sub_account(
    client_redis: object,
    orders_cached: list,
    positions_cached: list,
    query_trades: str,
    subaccounts_details_result: list,
    sub_account_cached_channel: str,
    message_byte_data: dict,
) -> None:

    if subaccounts_details_result:

        open_orders = [o["open_orders"] for o in subaccounts_details_result]

        if open_orders:
            caching.update_cached_orders(
                orders_cached,
                open_orders[0],
                "rest",
            )
        positions = [o["positions"] for o in subaccounts_details_result]

        if positions:
            caching.positions_updating_cached(
                positions_cached,
                positions[0],
                "rest",
            )

    my_trades_active_all = await db_mgt.executing_query_with_return(query_trades)

    data = dict(
        positions=positions_cached,
        open_orders=orders_cached,
        my_trades=my_trades_active_all,
    )

    message_byte_data["params"].update({"channel": sub_account_cached_channel})
    message_byte_data["params"].update({"data": data})

    await redis_client.publishing_result(
        client_redis,
        sub_account_cached_channel,
        message_byte_data,
    )
